import collections
import json

from valley.exceptions import ValidationException


class BaseSchema(object):
    """
    Base class for all Valley Schema classes.
    """
    _is_valid = False
    _create_error_dict = False

    def __init__(self, **kwargs):
        self._data = self.process_schema_kwargs(kwargs)
        self._errors = {}

    def __repr__(self):
        return '<{class_name}: {uni} >'.format(
            class_name=self.__class__.__name__, uni=self.__unicode__())

    def __unicode__(self):
        return '({0} Object)'.format(self.__class__.__name__)

    def __getattr__(self, name):
        if name in list(self._base_properties.keys()):
            prop = self._base_properties[name]
            return prop.get_python_value(self._data.get(name))

    def __setattr__(self, name, value):
        if name in list(self._base_properties.keys()):
            self._data[name] = value
        else:
            super(BaseSchema, self).__setattr__(name, value)

    def process_schema_kwargs(self, kwargs):
        schema_obj = {}
        for key, prop in list(self._base_properties.items()):
            try:
                value = prop.get_python_value(kwargs.get(key) or prop.get_default_value())
            except ValueError:
                value = kwargs.get(key) or prop.get_default_value()

            schema_obj[key] = value
        for i in self.BUILTIN_DOC_ATTRS:
            if kwargs.get(i):
                schema_obj[i] = kwargs[i]
        return schema_obj

    def validate(self):
        data = self._data.copy()
        for key, prop in list(self._base_properties.items()):
            prop_validate = getattr(self, '{}_validate'.format(key))
            try:
                prop.validate(data.get(key), key)
                # This allows devs to specify additional validation for a property
                if isinstance(prop_validate, collections.Callable):
                    prop_validate(data.get(key))
            except ValidationException as e:

                if self._create_error_dict:
                    self._errors[key] = e.error_msg
                else:
                    raise e
            value = prop.get_python_value(data.get(key))
            data[key] = value
        if self._create_error_dict and len(self._errors) < 1:
            self._is_valid = True
        else:
            self._is_valid = False
        self.cleaned_data = data

    @classmethod
    def get_class_name(cls):
        return cls.__name__.lower()

    def to_json(self):
        return json.dumps(self._data)

    def to_dict(self):
        return self._data



import importlib


def import_util(imp):
    '''
    Lazily imports a utils (class,
    function,or variable) from a module) from
    a string.
    @param imp:
    '''

    mod_name, obj_name = imp.rsplit('.', 1)
    mod = importlib.import_module(mod_name)
    return getattr(mod, obj_name)





import unittest
import json

from valley.utils import import_util
from valley.utils.json_utils import (ValleyEncoder,ValleyDecoder,
                                     ValleyEncoderNoType)
from valley.tests.examples.schemas import durham


class UtilTest(unittest.TestCase):

    json_string = '{"dogs": [{"breed": {"name": "Cocker Spaniel",' \
                      ' "_type": "valley.tests.examples.schemas.Breed"},' \
                      ' "name": "Bruno", "_type": "valley.tests.examples.' \
                      'schemas.Dog"}, {"breed": {"name": "Cockapoo", ' \
                      '"_type": "valley.tests.examples.schemas.Breed"}' \
                      ', "name": "Blitz", "_type": ' \
                      '"valley.tests.examples.schemas.Dog"}], ' \
                      '"primary_breed": {"name": "Cocker Spaniel", ' \
                      '"_type": "valley.tests.examples.schemas.Breed"}, ' \
                      '"name": "Durham", "_type": ' \
                      '"valley.tests.examples.schemas.Troop"}'

    def test_import_util(self):
        klass = import_util('valley.mixins.VariableMixin')
        self.assertEqual('VariableMixin', klass.__name__)

    def test_json_encoder(self):
        self.assertEqual(json.dumps(durham,cls=ValleyEncoder),self.json_string)

    def test_json_decoder(self):
        new_troop = json.loads(self.json_string, cls=ValleyDecoder)
        self.assertEqual(new_troop.name,durham.name)
        self.assertEqual(new_troop.primary_breed.name, durham.primary_breed.name)
        self.assertEqual(new_troop.dogs[0].name, durham.dogs[0].name)
        self.assertEqual(new_troop.dogs[1].name, durham.dogs[1].name)



from .properties import (BaseProperty,BooleanProperty,CharProperty,
                         DateProperty,DateTimeProperty,DictProperty,
                         EmailProperty,FloatProperty,IntegerProperty,
                         ListProperty,SlugProperty,
                         )
from .schema import BaseSchema


import json
import inspect

from .imports import import_util


class ValleyEncoder(json.JSONEncoder):
    show_type = True

    def default(self, obj):
        if not isinstance(obj, (list,dict,int,float,bool)):
            obj_dict = obj.to_dict()
            if self.show_type:
                obj_dict['_type'] = '{}.{}'.format(inspect.getmodule(obj).__name__, obj.__class__.__name__)
            return obj_dict
        return super(ValleyEncoder, self).default(obj)


class ValleyEncoderNoType(ValleyEncoder):
    show_type = False


class ValleyDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):
        if '_type' not in obj:
            return obj
        klass = import_util(obj['_type'])
        obj.pop('_type')

        return klass(**obj)



import collections
import datetime
import json
import time

from valley.exceptions import ValidationException
from valley.utils.json_utils import ValleyEncoder
from .validators import (RequiredValidator, StringValidator,
                         MaxLengthValidator, MinLengthValidator,
                         IntegerValidator, MaxValueValidator,
                         MinValueValidator, FloatValidator,
                         DateValidator, DateTimeValidator, BooleanValidator,
                         SlugValidator, EmailValidator, DictValidator,
                         ChoiceValidator, ListValidator, ForeignValidator,
                         ForeignListValidator, MultiValidator)


class VariableMixin(object):

    def get_validators(self):
        if self.required:
            self.validators.insert(0, RequiredValidator())
        if self.choices:
            self.validators.insert(0, ChoiceValidator(self.choices))

    def validate(self, value, key):
        if not value and not isinstance(self.get_default_value(), type(None)):
            value = self.get_default_value()
        for i in self.validators:
            i.validate(value, key)

    def get_default_value(self):
        """ return default value """
        default = self.default_value
        if isinstance(default, collections.Callable):
            default = default()
        return default

    def get_db_value(self, value):
        return value

    def get_python_value(self, value):
        return value


class CharVariableMixin(VariableMixin):

    def get_validators(self):
        super(CharVariableMixin, self).get_validators()
        self.validators.append(StringValidator())
        if self.kwargs.get('min_length'):
            self.validators.append(MinLengthValidator(
                self.kwargs.get('min_length')))

        if self.kwargs.get('max_length'):
            self.validators.append(MaxLengthValidator(
                self.kwargs.get('max_length')))

    def get_db_value(self, value):
        return str(value)

    def get_python_value(self, value):
        if not value:
            return None
        return str(value)

class SlugVariableMixin(CharVariableMixin):

    def get_validators(self):
        super(SlugVariableMixin, self).get_validators()
        self.validators.append(SlugValidator())


class EmailVariableMixin(CharVariableMixin):

    def get_validators(self):
        super(EmailVariableMixin, self).get_validators()
        self.validators.append(EmailValidator())


class NumericVariableMixin(VariableMixin):

    def get_validators(self):
        super(NumericVariableMixin, self).get_validators()

        if self.kwargs.get('max_value'):
            self.validators.append(MaxValueValidator(
                self.kwargs.get('max_value')))

        if self.kwargs.get('min_value'):
            self.validators.append(MinValueValidator(
                self.kwargs.get('min_value')))


class IntegerVariableMixin(NumericVariableMixin):

    def get_validators(self):
        super(IntegerVariableMixin, self).get_validators()
        self.validators.insert(0, IntegerValidator())

    def get_db_value(self, value):
        return int(value)

    def get_python_value(self, value):
        if not value:
            return None
        return int(value)


class FloatVariableMixin(NumericVariableMixin):

    def get_validators(self):
        super(FloatVariableMixin, self).get_validators()
        self.validators.insert(0, FloatValidator())

    def get_db_value(self, value):
        return float(value)

    def get_python_value(self, value):
        if not value:
            return None
        return float(value)


class DateMixin(VariableMixin):

    def get_validators(self):
        super(DateMixin, self).get_validators()
        self.validators.insert(0, DateValidator())

    def now(self):
        return datetime.datetime.now().date()

    def get_default_value(self):
        default = self.default_value
        if self.auto_now or self.auto_now_add:
            return self.now()
        return default

    def get_python_value(self, value):
        if not value:
            return None
        if isinstance(value, str):
            try:
                value = datetime.date(*time.strptime(value, '%Y-%m-%d')[:3])
            except ValueError as e:
                raise ValueError('Invalid ISO date %r [%s]' % (value,
                                                               str(e)))
        return value

    def get_db_value(self, value):
        if value is None:
            return value
        return value.isoformat()


class DateTimeMixin(VariableMixin):

    def get_validators(self):
        super(DateTimeMixin, self).get_validators()
        self.validators.insert(0, DateTimeValidator())

    def get_default_value(self):
        default = self.default_value
        if self.auto_now or self.auto_now_add:
            return self.now()
        return default

    def get_python_value(self, value):
        if isinstance(value, str):
            try:
                value = value.split('.', 1)[0]  # strip out microseconds
                value = value[0:19]  # remove timezone
                value = datetime.datetime.strptime(value, '%Y-%m-%dT%H:%M:%S')
            except ValueError as e:
                raise ValueError('Invalid ISO date/time %r [%s]' %
                                 (value, str(e)))
        return value

    def get_db_value(self, value):
        if not value:
            return None
        if self.auto_now:
            value = self.now()

        if value is None:
            return value

        return value.replace(microsecond=0).isoformat() + 'Z'

    def now(self):
        return datetime.datetime.utcnow()


class BooleanMixin(VariableMixin):

    def get_validators(self):
        self.validators = [BooleanValidator()]

    def get_db_value(self, value):
        return bool(value)

    def get_python_value(self, value):
        true_vals = ('True', 'true', 1, '1',True)
        false_vals = ('False', 'false', 0, '0',False)
        if value in true_vals:
            value = True
        elif value in false_vals:
            value = False
        else:
            raise ValueError('This value is not a boolean value.')
        return value


class DictMixin(VariableMixin):

    def get_validators(self):
        super(DictMixin, self).get_validators()
        self.validators.insert(0, DictValidator())

    def get_db_value(self, value):
        return json.dumps(value)

    def get_python_value(self, value):
        if not value:
            return None
        try:
            return json.loads(value)
        except TypeError:
            return dict(value)


class ListMixin(VariableMixin):

    def get_validators(self):
        super(ListMixin, self).get_validators()
        self.validators.insert(0, ListValidator())

    def get_db_value(self, value):
        return json.dumps(value)

    def get_python_value(self, value):
        if not value:
            return None
        try:
            return json.loads(value)
        except TypeError:
            try:
                return list(value)
            except TypeError:
                return value


class ForeignMixin(VariableMixin):

    def get_validators(self):
        super(ForeignMixin, self).get_validators()
        self.validators.insert(0, ForeignValidator(self.foreign_class))

    def get_db_value(self, value):
        if self.return_type == 'single':
            if not self.return_prop:
                raise ValueError('ForeignProperty classes requires the '
                    'return_prop argument if return_type equals "single"')
            return value._data[self.return_prop]
        if self.return_type == 'dict':
            return value._data
        if self.return_type == 'json':
            return json.dumps(value, cls=ValleyEncoder)
        else:
            return value


class ForeignListMixin(ListMixin):

    def get_validators(self):
        super(ForeignListMixin, self).get_validators()
        self.validators.insert(len(self.validators),ForeignListValidator(self.foreign_class))

    def get_db_value(self, value):
        if self.return_type == 'single':
            if not self.return_prop:
                raise ValueError('ForeignProperty classes requires the '
                    'return_prop argument if return_type equals "single"')
            return value._data[self.return_prop]
        if self.return_type == 'list':
            return [obj._data for obj in value]
        if self.return_type == 'json':
            return json.dumps(value, cls=ValleyEncoder)
        else:
            return value


class MultiMixin(VariableMixin):

    def get_validators(self):
        super(MultiMixin, self).get_validators()
        self.validators = [MultiValidator(self.validators)]




import unittest



from valley.exceptions import ValidationException
from valley.tests.examples.schemas import StudentB, Student, durham, Troop, bruno, blitz, cocker


class SchemaTestCase(unittest.TestCase):

    def setUp(self):
        self.student = Student(name='Frank White',slug='frank-white',
                    email='frank@white.com',age=18,
                    gpa=3.0,date='2017-01-10',
                    datetime='2017-01-10T12:00:00',active=False
                    )

        self.troop = Troop(name='Durham',dogs=[bruno,blitz],
                           primary_breed=cocker)

    def test_valid(self):
        self.student.validate()
        self.assertDictEqual(self.student._errors, {})

    def test_foreign_valid(self):
        self.troop.validate()
        self.assertDictEqual(self.troop._errors,{})

    def test_foreign_property_wrong_type(self):
        self.troop.primary_breed = 'Cocker'
        self.troop.validate()
        self.assertDictEqual({'primary_breed': 'primary_breed: This value '
                                               '(Cocker) should be an instance of Breed.'},
                             self.troop._errors)

    def test_foreign_list_property_wrong_type(self):
        self.troop.dogs = ['Test',bruno,blitz]
        self.troop.validate()
        self.assertDictEqual({'dogs': 'dogs: This value (Test) should '
                                      'be an instance of Dog.'},
                             self.troop._errors)

    def test_long_name(self):
        self.student.name = 'Frank Lindsay Hightower III'
        self.student.validate()
        ed = {'name': 'name: This value should have a length lesser than or equal to 20. '
                      'Currently Frank Lindsay Hightower III'}
        self.assertDictEqual(ed, self.student._errors)

    def test_short_name(self):
        self.student.name = 'Ira'
        self.student.validate()
        ed = {'name': 'name: This value should have a length greater '
                      'than or equal to 5. Currently Ira'}
        self.assertDictEqual(ed,self.student._errors)

    def test_no_name(self):
        self.student.name = None
        self.student.validate()
        ed = {'name': 'name: This value is required'}
        self.assertDictEqual(ed, self.student._errors)

    def test_slug_space(self):
        self.student.slug = 'Some City'
        self.student.validate()
        ed = {'slug': 'slug: This value should be a slug. ex. pooter-is-awesome'}
        self.assertDictEqual(ed, self.student._errors)

    def test_email_space(self):
        self.student.email = 'Some City'
        self.student.validate()
        ed = {'email': 'email: This value should be a valid email address'}
        self.assertDictEqual(ed, self.student._errors)

    def test_email_wrong(self):
        self.student.email = 'e+ e@g.com'
        self.student.validate()
        ed = {'email': 'email: This value should be a valid email address'}
        self.assertDictEqual(ed, self.student._errors)

    def test_age_numeric_string(self):
        self.student.age = '5'
        self.student.validate()
        ed = {'age': 'age: This value should be an integer'}
        self.assertEqual(ed, self.student._errors)

    def test_age_numeric_float(self):
        self.student.age = 5.0
        self.student.validate()
        ed = {'age': 'age: This value should be an integer'}
        self.assertDictEqual(ed, self.student._errors)


class SchemaMethods(unittest.TestCase):

    def setUp(self):
        attr_dict = {
            'name':'Frank White',
            'slug':'frank-white',
            'email':'frank@white.com',
            'age':18, 'gpa':3.0,
            'date':'2017-01-10',
            'datetime':'2017-01-10T12:00:00'
        }
        self.student = Student(**attr_dict)
        self.studentb = StudentB(**attr_dict)

    def test_setattr(self):
        self.student.name = 'Curious George'
        self.assertEqual(self.student._data.get('name'),'Curious George')

    def test_getattr(self):
        self.assertEqual(self.student.name,'Frank White')

    def test_process_doc_kwargs(self):
        response = self.student.process_schema_kwargs({'name':1,'slug':1})
        self.assertEqual(response,{'name':'1','slug':'1','email':None,
                                   'age':None,'gpa':None,'date':None,
                                   'active':False,
                                   'datetime':None})

    def test_error_dict_false_validate(self):
        self.studentb.name = 1
        self.assertRaises(ValidationException,self.studentb.validate)




from setuptools import setup, find_packages


def parse_requirements(filename):
    """ load requirements from a pip requirements file """
    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith("#")]

version = '1.5.2'

LONG_DESCRIPTION = """
=======================
Valley
=======================

Python extensible schema validations and declarative syntax helpers.


"""

setup(
    name='valley',
    version=version,
    description="""Python extensible schema validations and declarative syntax helpers.""",
    long_description=LONG_DESCRIPTION,
    classifiers=[
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Environment :: Web Environment",
    ],
    keywords='validations,schema,declarative',
    author='Brian Jinwright',
    author_email='opensource@capless.io',
    maintainer='Brian Jinwright',
    packages=find_packages(),
    url='https://github.com/capless/valley',
    license='GNU GPL V3',
    install_requires=parse_requirements('requirements.txt'),
    include_package_data=True,
)



import logging

from valley.mixins import VariableMixin, CharVariableMixin, \
    IntegerVariableMixin, FloatVariableMixin, BooleanMixin, \
    DateMixin, DateTimeMixin, SlugVariableMixin, EmailVariableMixin, \
    DictMixin, ListMixin, ForeignMixin, ForeignListMixin, MultiMixin


LOGGER = logging.getLogger('valley')


class BaseProperty(VariableMixin, object):
    default_value = None
    allow_required = True
    def __init__(
        self,
        default_value=None,
        required=False,
        validators=[],
        choices=None,
        verbose_name=None,
        **kwargs
    ):
        self.default_value = default_value or self.default_value
        if self.allow_required == True:
            self.required = required
        else:
            self.required = False
            if required == True:
                LOGGER.warning(
                    'The required argument has no effect on {}'.format(
                        self.__class__.__name__))
        self.choices = choices
        self.kwargs = kwargs
        if isinstance(validators,list):
            self.validators = list(set(validators))
        else:
            self.validators = []

        self.get_validators()
        self.validators = set(self.validators)

        if verbose_name:
            self.verbose_name = verbose_name


class CharProperty(CharVariableMixin, BaseProperty):
    pass


class SlugProperty(SlugVariableMixin, CharProperty):
    pass


class EmailProperty(EmailVariableMixin, CharProperty):
    pass


class IntegerProperty(IntegerVariableMixin, BaseProperty):
    pass


class FloatProperty(FloatVariableMixin, BaseProperty):
    pass


class BooleanProperty(BooleanMixin, BaseProperty):
    default_value = False
    allow_required = False

class DateProperty(DateMixin, BaseProperty):

    def __init__(
            self,
            default_value=None,
            required=True,
            validators=[],
            verbose_name=None,
            auto_now=False,
            auto_now_add=False,
            **kwargs):

        super(
            DateProperty,
            self).__init__(
            default_value=default_value,
            required=required,
            validators=validators,
            verbose_name=verbose_name,
            **kwargs)
        self.auto_now = auto_now
        self.auto_now_add = auto_now_add



class DateTimeProperty(DateTimeMixin, BaseProperty):

    def __init__(
            self,
            default_value=None,
            required=True,
            validators=[],
            verbose_name=None,
            auto_now=False,
            auto_now_add=False,
            **kwargs):

        super(
            DateTimeProperty,
            self).__init__(
            default_value=default_value,
            required=required,
            validators=validators,
            verbose_name=verbose_name,
            **kwargs)
        self.auto_now = auto_now
        self.auto_now_add = auto_now_add


class DictProperty(DictMixin, BaseProperty):
    pass


class ListProperty(ListMixin, BaseProperty):
    pass


class ForeignProperty(ForeignMixin, BaseProperty):

    def __init__(self,foreign_class,return_type=None,return_prop=None,**kwargs):
        self.foreign_class = foreign_class
        super(ForeignProperty, self).__init__(**kwargs)
        self.return_type = return_type
        self.return_prop = return_prop


class ForeignListProperty(ForeignListMixin, BaseProperty):

    def __init__(self,foreign_class,return_type=None,return_prop=None,**kwargs):
        self.foreign_class = foreign_class
        super(ForeignListProperty, self).__init__(**kwargs)
        self.return_type = return_type
        self.return_prop = return_prop


class MultiProperty(MultiMixin,BaseProperty):

    def validate(self, value, key):
        super(MultiProperty, self).validate(value,key)
